"""Storage module for class GetterXML"""

import requests
import logging

logger = logging.getLogger('app.getter')


class GetterXml:
    """Class for receiving response with xml from the server"""

    def __init__(self):
        self.response = None
        self.url = None

    def get_response(self, url: str):
        """
        Get response and check status.
        :param url: source url for request
        :return: response
        """
        self.url = url

        logger.info(f'Make request to URL - {self.url}')

        try:
            self.response = requests.get(self.url)
            logger.debug(self.response)
        except ConnectionError as err:
            logger.error(err)

        if self.response.ok:
            if GetterXml._is_response_status_200(self.response):
                logger.info(f'Status response is good')
                return self.response
            else:
                logger.warning(f'Current status code response is {self.response.status_code}')
                return self.response

    @staticmethod
    def _is_response_status_200(resp) -> bool:
        """
        Check is status code 200
        :param resp: response object
        :return: bool
        """
        if resp.status_code == 200:
            return True
        return False



"""Main module for working with APP"""

import os
import sys
from pathlib import Path

import logging

sys.path.append(os.path.dirname(os.getcwd()))

from grebarss_reader.config import AppLogger
from grebarss_reader.grebarss_parser import RssParser

AppLogger.init_logger('app')
logger = logging.getLogger("app.rss_reader")


def main():
    """Starts APP"""
    rss_parser = RssParser()
    try:
        rss_parser.start()
    except Exception as exc_obj:
        logger.exception(f"Rss reader crashed from {exc_obj}")
        print(f"Rss reader crashed from {type(exc_obj).__name__}. For more information see log file.")


if __name__ == "__main__":
    logger.info('RSS reader started working.')
    main()
    logger.info('RSS reader finished working.')



"""Storage module with settings for argparse and logging"""

import logging
import sys
from enum import Enum
from pathlib import Path

import argparse


class AppConstant(Enum):
    """Class immutable values"""
    ACTUAL_VERSION = 'Version 4.1.0'


class AppArgParser:
    """Class for initializing arguments for working with CLI"""

    def __init__(self):
        self.parser = argparse.ArgumentParser(description='Pure Python command-line RSS reader.')
        self.parser.add_argument('--version', action="version", version=AppConstant.ACTUAL_VERSION.value,
                                 help='Print version info')
        self.parser.add_argument('--json', action='store_true', help='Print result as JSON in stdout')
        self.parser.add_argument('--verbose', action="store_true", help='Outputs verbose status messages')
        self.parser.add_argument('--limit', type=int, default=None, help='Limit news topics if this parameter provided')
        self.parser.add_argument('source', nargs="?", default=None, help='URL RSS')
        self.parser.add_argument('--date', type=int, default=None,
                                 help='Print news published on a specific date from cache')
        self.parser.add_argument('--to-html', type=str, nargs="?", metavar="PATH",
                                 help="Convert news to HTML format and save them by the specified folder path")
        self.parser.add_argument('--to-pdf', type=str, nargs="?", metavar="PATH",
                                 help="Convert news to PDF format and save them by the specified folder path")

    def get_args(self) -> argparse.Namespace:
        """
        Initialization of arguments
        :return: object storing attributes
        """
        return self.parser.parse_args()


class AppLogger:
    """Class for initialization and setup logger and handlers"""

    FORMAT = '%(asctime)s - %(name)s:%(lineno)s - %(levelname)s - %(message)s'

    @staticmethod
    def init_logger(name):
        """Initialization and setup root logger. Setup and start file handler"""
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        current_dir = Path(__file__).parent.resolve()
        file_path = Path(current_dir, 'logs/grebarss_logs.log')

        fh = logging.FileHandler(filename=file_path, mode='w', encoding='utf-8')
        fh.setFormatter(logging.Formatter(AppLogger.FORMAT))
        fh.setLevel(logging.DEBUG)
        logger.addHandler(fh)

    @staticmethod
    def activate_verbose():
        """Setup and start stream handler for verbose mode"""
        logger = logging.getLogger('app')
        sh = logging.StreamHandler(stream=sys.stdout)
        sh.setFormatter(logging.Formatter(AppLogger.FORMAT))
        sh.setLevel(logging.INFO)
        logger.addHandler(sh)



"""Storage module for classes RssParser"""

import os
import sys
import json

import logging
import re

import xmltodict

sys.path.append(os.path.dirname(os.getcwd()))

import grebarss_reader.urlmarker as urlmarker
import grebarss_reader.config as config
from grebarss_reader.cacher import Cacher
from grebarss_reader.converter import Converter
from grebarss_reader.getter import GetterXml
from grebarss_reader.printer import Printer

logger = logging.getLogger('app.rss_parser')


class RssParser:
    """Class for parsing response received by getter.py and printing result"""

    @staticmethod
    def _get_image_link(item: dict) -> str:
        url_list = set(re.findall(urlmarker.URL_REGEX, str(item)))
        logger.debug(f'{type(url_list)}Url list - {url_list}')
        for link in url_list:
            if link.endswith('.jpg') or link.endswith('.jpeg') or link.endswith('.png') or link.startswith(
                    'https://s.yimg') or link.startswith('https://i.guim'):
                return link

    @staticmethod
    def _get_raw_feed(source: str):
        """
        Transform XML data to dict with all news
        :arg source: URL source with news
        :return: dictionary with raw feed (all news)
        """
        raw_feed = xmltodict.parse(GetterXml().get_response(source).text, encoding='utf-8')
        logger.debug(f'All news in raw dict (feed) - {raw_feed}')
        return raw_feed

    @staticmethod
    def _parse_xml(raw_feed) -> dict:
        """
        Transform XML data to dict
        :arg args: set of arguments
        :return: dictionary with XLM data
        """
        data_dict_out = {"item": []}

        for item in raw_feed['rss']['channel']['item']:
            data_dict_out['item'].append(
                {"title": item.get("title"),
                 "pubDate": item.get("pubDate"),
                 "link": item.get("link"),
                 "image": RssParser._get_image_link(item)})

        logger.debug(f'All news in final dict - {data_dict_out}')
        return data_dict_out

    @staticmethod
    def _limit(data: dict, limit: int) -> dict:
        out_dict = dict()
        out_dict['item'] = (data['item'][:limit])
        return out_dict

    @staticmethod
    def start():
        """ Start work with rss_parser"""
        logger.info("Module rss_parser is starting.")

        args = config.AppArgParser().get_args()
        logger.debug(f'Argparse sent these arguments - {args.__dict__}')

        if args.verbose:
            config.AppLogger.activate_verbose()
            logger.info(f'Verbose mode activated.')

        parser = RssParser()

        if args.date:
            cache: dict = Cacher(args.source).get_cache_data(args.date)
            limited_cache = RssParser._limit(cache, args.limit)
            logger.debug(f'limited cache - {limited_cache}')
            if args.to_html:
                Converter(args.source, limited_cache, args.to_html).to_html()
            elif args.to_pdf:
                Converter(args.source, limited_cache, args.to_pdf).to_pdf()
            else:
                Printer().print_info(limited_cache)
        else:
            feed: dict = parser._parse_xml(parser._get_raw_feed(args.source))
            limited_feed = RssParser._limit(feed, args.limit)
            logger.debug(f'limited feed for cache - {limited_feed}')
            Cacher(args.source).cache(limited_feed)

            if args.json:
                logger.info(f'Json mode activated.')
                Printer().print_info_json(limited_feed)
                if args.to_html:
                    Converter(args.source, limited_feed, args.to_html).to_html()
                elif args.to_pdf:
                    Converter(args.source, limited_feed, args.to_pdf).to_pdf()
            else:
                if args.to_html:
                    Converter(args.source, limited_feed, args.to_html).to_html()
                elif args.to_pdf:
                    Converter(args.source, limited_feed, args.to_pdf).to_pdf()
                else:
                    Printer().print_info(limited_feed)



"""Storage module for classes Printer"""

import json
import logging

logger = logging.getLogger('app.printer')


class Printer:
    """Class for printing info"""

    @staticmethod
    def print_info(data_to_print: dict):
        """
        Print info in stdout
        :param data_to_print: data for printing
        """

        for item in data_to_print['item']:
            print('\n', "- " * 20, '\n')
            print(f'Title: {item.get("title")}\nData: {item.get("pubDate")}\nLink: {item.get("link")}\n'
                  f'Image: {item.get("image")}')
            if item.get('source'):
                print(f'Source: {item.get("source")}')

    @staticmethod
    def print_info_json(data_to_print: dict):
        """
        Print info in json format
        :param data_to_print: data for printing
        """

        for item in data_to_print['item']:
            item_in_json = json.dumps(item, ensure_ascii=False).encode('utf8')
            print('\n', "- " * 20, '\n')
            print(item_in_json.decode())



"""Storage module for Converter class"""

import logging
from pathlib import Path

from jinja2 import Template
from xhtml2pdf import pisa

logger = logging.getLogger('app.converter')


class Converter:
    """
    Class for converting news into HTML or PDF received from cacher (--date) or rss_parser (all other cases)"""

    def __init__(self, source, feed: dict, folder_path: str):
        self.folder_path = folder_path
        self.feed = feed
        self.source = source
        """
        :param source: URL of the RSS source. If not specifying RSS source - None
        :param feed: dict with limited news
        :param folder_path: path to the folder where the converted files will be created
        """

    @staticmethod
    def _get_html_with_temp(**kwargs) -> str:
        """
        Gets html string using a template for further conversion
        :param kwargs: data to install into the template
        :return: html str with data
        """
        current_dir = Path(__file__).parent.resolve()
        path_to_templates = Path(current_dir, "templates")

        template = Template(open(Path(path_to_templates, "html_template.jinja2")).read())
        return template.render(**kwargs)

    def to_html(self):
        """Create file with converted feed to HTML format"""
        file_path = Path(self.folder_path, "news.html")

        try:
            logger.info("Start converting feeds to HTML")
            with open(file_path, 'w', encoding='utf-8') as html_file:
                html_file.write(self._get_html_with_temp(
                    feed=self.feed,
                    url=self.source,
                    fonts_path=str(Path(Path(__file__).parent.resolve(), "fonts")),
                ))
        except Exception as exc:
            logger.exception(exc)
        else:
            logger.info(f"HTML saved in {file_path}")

    def to_pdf(self):
        """Create file with converted feed to PDF format"""
        file_path = Path(self.folder_path, "news.pdf")

        try:
            logger.info("Start converting feeds to PDF")
            with open(file_path, 'w+b') as pdf_file:
                pisa.CreatePDF(self._get_html_with_temp(
                    feed=self.feed,
                    url=self.source,
                    fonts_path=str(Path(Path(__file__).parent.resolve(), "fonts")),
                ), dest=pdf_file)
        except Exception as exc:
            logger.exception(exc)
        else:
            logger.info(f"PDF saved in {file_path}")



import os

from setuptools import find_packages, setup


def read(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as file:
        return file.read()


setup(
    name="grebarss",
    version="4.1.11",
    author="Alexander Hreben",
    author_email="greba3000@gmail.com",
    url="https://github.com/Greba3000/Homework",
    # description="CLI utility for reading news",
    # long_description=read("README.md"),
    python_requires=">=3.9",
    classifiers=["Programming Language :: Python :: 3.9"],
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=read("requirements.txt").splitlines(),
    entry_points={
        "console_scripts": [
            "grebarss=grebarss_reader.rss_reader:main",
            "rss_reader-reader=grebarss_reader.rss_reader:main",
        ],
    },
)


"""Storage module for Cacher class"""

from pathlib import Path

import json
import logging
from datetime import datetime

logger = logging.getLogger('app.cacher')


class Cacher:
    """
    Class for saving and retrieving news from the repository (db.json)
    if the variable source == None it means that we use all the sources available in the storage (db.json)
    """

    def __init__(self, source):
        self.source = source
        current_dir = Path(__file__).parent.resolve()
        file_path = Path(current_dir, 'database/db.json')
        self.cache_file_path = file_path
        """
        :param source: URL of the RSS resource
        """

    def cache(self, feed: dict):
        """
        Saves news from rss_parser to the repository (db.json)
        :param feed: dictionary containing news from rss_parser
        """
        logger.info('Module cacher is starting to save news.')

        with open(self.cache_file_path, "r+", encoding="utf-8") as cache_file:
            json_content: str = cache_file.read()
            json_dicts: list = json.loads(json_content)

            url_list = []
            counter = 0

            for elem in json_dicts:
                url_list.append([key for key in elem.keys()])
                if self.source in elem:
                    for item in feed.get('item'):
                        if item not in json_dicts[counter].get(self.source)[0].get('item'):
                            json_dicts[counter].get(self.source)[0].get('item').append(item)
                counter += 1

            logger.debug(f'URL used list - {url_list}')
            if [self.source] not in url_list:
                json_dicts.append({self.source: [feed]})
            logger.debug(f'News for save - {type(json_dicts)}, {json_dicts}')

        with open(self.cache_file_path, "w", encoding="utf-8") as cache_file:
            json.dump(json_dicts, cache_file, indent=4, ensure_ascii=False)
        logger.info('Module cacher is finishing.')

    def get_cache_data(self, date: int) -> dict:
        """
        Creates dictionary of news from the repository (db.json) published on a specific date.
        :param date: value date in %Y%m%d format
        :return: dict containing news from repository (db.json)
        """
        out_dict_one_source = {'item': []}
        out_dict_all_sources = {'item': []}

        with open(self.cache_file_path, "r", encoding="utf-8") as cache_file:
            json_dicts: list = json.loads(cache_file.read())

            if self.source:
                counter_dict_source_pos = 0
                for elem in json_dicts:
                    if self.source in elem:
                        for item in json_dicts[counter_dict_source_pos].get(self.source)[0].get('item'):
                            if date == Cacher._get_convert_date(item['pubDate']):
                                out_dict_one_source.get('item').append(item)
                    counter_dict_source_pos += 1
                logger.debug(f'Out dict with source news from cache - {out_dict_one_source}')
                logger.info('Module cacher is finishing.')
                return out_dict_one_source
            else:
                counter_dict_all_pos = 0
                for elem in json_dicts:
                    for key_url in elem.keys():
                        for item in elem.get(key_url)[0].get('item'):
                            if date == Cacher._get_convert_date(item['pubDate']):
                                item['source'] = key_url
                                out_dict_all_sources.get('item').append(item)
                    counter_dict_all_pos += 1
                logger.debug(f'Out dict with all news from cache - {out_dict_all_sources}')
                logger.info('Module cacher is finishing.')
                return out_dict_all_sources

    @staticmethod
    def _get_convert_date(date: str) -> int:
        """
        Convert date to %Y%m%d format
        :param date: date in RFC 822 format
        :return: parsed date in %Y%m%d format
        """
        allowed_date_formats = [
            "%a, %d %b %Y %H:%M:%S %z",
            "%a, %d %b %Y %H:%M:%S %Z",
            "%Y-%m-%dT%H:%M:%SZ",
        ]
        temp_date = None

        for date_format in allowed_date_formats:
            try:
                temp_date = datetime.strptime(date, date_format)
            except ValueError:
                pass

        if not temp_date:
            raise ValueError(f'{date} - incorrect data format')
        return int(f"{temp_date.year}{temp_date.month:02d}{temp_date.day:02d}")
